/*
 *  Memory management and utility functions for the
 *  Software Engineering course.
 *
 *  Written by: Loki  2000/10/20  version 9
 *  Modified by: Andrew Grimm (added memory-allocation reporting).
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include "zutil.h"

/*
 *  Internal definitions.
 */
#define USED_START 0x5AFEC0DE
#define USED_END   0xC0CAC01A
#define FREE_START 0xDEADD0D0
#define FREE_END   0xBADD15C0
#define FILL_BYTE  0x55
#define MAX_BLOCK  (1<<27)

/*
 *  Global variables.
 *
 *  Comment out references to these globals if you want this code
 *  to be thread-safe. Globals prevent threads from working properly
 *  because two threads might access the same variable at the same time.
 */
static long mem_allocated = 0;
static long blocks_allocated = 0;
static long smallest_block = -1;
static long largest_block = -1;

/*
 *  Report memory usage to stderr.
 */
void zreport(void)
{
  fprintf(stderr, "zreport: total memory allocated     = %ld bytes\n",
	  mem_allocated);

  fprintf(stderr, "zreport: number of blocks allocated = %ld\n",
	  blocks_allocated);

  if (blocks_allocated > 0)
    fprintf(stderr, "zreport: average block size         = %ld bytes\n",
	    mem_allocated / blocks_allocated);

  if (smallest_block > -1)
    fprintf(stderr, "zreport: smallest block ever used   = %ld bytes\n",
	    smallest_block);

  if (largest_block > -1)
    fprintf(stderr, "zreport: largest block ever used    = %ld bytes\n",
	    largest_block);

  fprintf(stderr, "zreport: ---------------------------------------\n");
}

/*
 *  Print a message on stderr. The format of this function is
 *  the same as printf.
 */
void zerror(char *fmt, ...)
{
  va_list v;

  va_start(v, fmt);
  vfprintf(stderr, fmt, v);
  va_end(v);
}

/*
 *  Check that a pointer is a validly allocated pointer returned from
 *  zalloc, and that its buffers are intact. This function aborts if
 *  there is something wrong. NULL is assumed to be a valid pointer.
 */
void zcheck(void *v)
{
  char * ptr;
  long nbytes, length;
  long i, start, end;

  if (v == NULL)
    return;
  ptr = (char *) v;

  start = *(long *)(ptr - sizeof(long));
  if (start == FREE_START) {
    zerror("zcheck: block has already been freed.\n");
    zerror(" Start buffer is 0x%-08.8lX\n", start);
    abort();
  }
  else if (start != USED_START) {
    zerror("zcheck: block start buffer is wrong.\n");
    zerror(" Start buffer is 0x%-08.8lX\n", start);
    abort();
  }

  nbytes = *(long *)(ptr - sizeof(long) * 2);
  if (nbytes < 0) {
    zerror("zcheck: block size is negative (size is corrupt?).\n");
    zerror(" Block length is %ld.\n", nbytes);
    zerror(" Start buffer is 0x%-08.8lX\n", start);
    abort();
  }
  else if (nbytes > MAX_BLOCK) {
    zerror("zcheck: block is too big (size is corrupt?).\n");
    zerror(" Block length is %ld.\n", nbytes);
    zerror(" Start buffer is 0x%-08.8lX\n", start);
    abort();
  }

  length = (nbytes / sizeof(long) + 1) * sizeof(long); /* round up */
  end = *(long *)(ptr + length);
  if (end == FREE_END) {
    zerror("zcheck: block buffers disagree.\n");
    zerror(" Block length is %ld.\n", nbytes);
    zerror(" Start buffer is 0x%-08.8lX.\n", start);
    zerror(" End buffer is 0x%-08.8lX.\n",  end);
    abort();
  }
  else if (end != USED_END) {
    zerror("zcheck: end of block was overwritten.\n");
    zerror(" Block length is %ld.\n", nbytes);
    zerror(" Start buffer is 0x%-08.8lX.\n", start);
    zerror(" End buffer is 0x%-08.8lX.\n",  end);
    abort();
  }
  for (i=nbytes; i < length; i++) {
    if (ptr[i] == FILL_BYTE)
      continue;
    zerror("zcheck: end buffer corrupted at offset %d.\n", i);
    zerror(" Block length is %ld.\n", nbytes);
    zerror(" Start buffer is 0x%-08.8lX.\n", start);
    zerror(" End buffer is 0x%-08.8lX.\n",  end);
    abort();
  }
}

/*
 *  Return the length in bytes of the usable portion of this block.
 *  A NULL pointer is defined to have a length of 0 bytes.
 *  This function performs a complete integrity check of the pointer.
 */
long zlength(void *v)
{
  char * ptr;
  long nbytes;

  if (v == NULL)
    return 0;

  zcheck(v);	/* check that the pointer is valid */

  ptr = (char *) v;

  nbytes = *(long *)(ptr - sizeof(long) * 2);
  return nbytes;
}

/*
 *  Return a block of memory for use by the program. The routine
 *  uses malloc to obtain the block of memory, and then surrounds
 *  the block with a buffer zone of unlikely 'magic numbers' which
 *  are checked when the block is freed, to ensure memory has not
 *  been overwritten accidentally.
 *
 *  The returned block is not initialised and so may contain random,
 *  data, and if malloc fails, this routine causes the program to abort,
 *  so you can assume zalloc always returns a valid pointer.
 *
 *  Allocating a zero-sized block returns NULL. The malloc function
 *  may or may not behave this way, so you should not get into the
 *  habit of allocating zero-sized blocks.
 */
void * zalloc(long nbytes)
{
  char * ptr;
  long length;

  if (nbytes == 0)
    return NULL;
  else if (nbytes < 0) {
    zerror("zalloc: requested size is negative (%ld)\n", nbytes);
    abort();
  }
  else if (nbytes > MAX_BLOCK) {
    zerror("zalloc: requested size is too big (%ld)\n", nbytes);
    abort();
  }
  length = (nbytes / sizeof(long) + 1) * sizeof(long); /* round up */
  ptr = (char *) malloc(sizeof(long) * 3 + length);
  if (ptr == NULL) {
    zerror("zalloc: malloc returned NULL pointer.\n");
    abort();
  }

  mem_allocated = mem_allocated + nbytes; /* Andrew Grimm */
  blocks_allocated++;
  if ((smallest_block == -1) || (smallest_block > nbytes))
    smallest_block = nbytes;
  if (largest_block < nbytes)
    largest_block = nbytes;

  *(long *)(ptr) = nbytes;
  *(long *)(ptr+sizeof(long)) = USED_START;
  ptr += sizeof(long) * 2;	/* point to user block */
  memset(ptr+nbytes, FILL_BYTE, length-nbytes);
  *(long *)(ptr+length) = USED_END;
  return ptr;
}

/*
 *  Realloc a block after checking the buffers are intact.
 *  Reallocing a NULL pointer is equivalent to using zalloc,
 *  and reallocing to a size of zero bytes is the same as
 *  using zfree.
 */
void * zrealloc(void *v, long nbytes)
{
  char *ptr;
  long int oldsize, length;
  
  if (v == NULL)
    return zalloc(nbytes);
  else if (nbytes == 0) {
    zfree(v);
    return NULL;
  }
  else if (nbytes < 0) {
    zerror("zrealloc: requested size is negative (%ld)\n", nbytes);
    abort();
  }
  else if (nbytes > MAX_BLOCK) {
    zerror("zrealloc: requested size is too big (%ld)\n", nbytes);
    abort();
  }

  zcheck(v);	/* check that the pointer is valid */

  ptr = ((char *) v) - sizeof(long) * 2;
  oldsize = *(long *)(ptr);
	
  length = (nbytes / sizeof(long) + 1) * sizeof(long); /* round up */
  ptr = (char *) realloc(ptr, sizeof(long) * 3 + length);
  if (ptr == NULL) {
    zerror("zrealloc: realloc returned NULL pointer.\n");
    abort();
  }

  mem_allocated = mem_allocated + nbytes - oldsize; /* Andrew Grimm */
  if (smallest_block > nbytes)
    smallest_block = nbytes;
  if (largest_block < nbytes)
    largest_block = nbytes;

  *(long *)(ptr) = nbytes;	/* USED_START already there */
  ptr += sizeof(long) * 2;	/* point to user block */
  memset(ptr+nbytes, FILL_BYTE, length-nbytes);
  *(long *)(ptr+length) = USED_END;
  return ptr;
}

/*
 *  Free the block of memory previously returned by zalloc.
 *  Checks that the block was created by zalloc, and has not already
 *  been freed. Checks that the boundaries have not been overwritten.
 *
 *  Before freeing the block, we write an unlikely value into the
 *  memory of the block and mark it as having been freed, to prevent
 *  mistaken reuse of this block afterwards.
 *
 *  This function will abort if there is a problem.
 *
 *  Freeing null pointers is allowed and does nothing.
 */
void zfree(void *v)
{
  char * ptr;
  long nbytes, length;
  long start, end;

  if (v == NULL)
    return;
  zcheck(v);	/* check that the pointer is valid */

  ptr = (char *) v;

  start = *(long *)(ptr - sizeof(long));
  nbytes = *(long *)(ptr - sizeof(long) * 2);
  length = (nbytes / sizeof(long) + 1) * sizeof(long); /* round up */
  end = *(long *)(ptr + length);

  mem_allocated = mem_allocated - nbytes; /* Andrew Grimm */
  blocks_allocated--;

  *(long *)(ptr - sizeof(long)) = FREE_START;
  *(long *)(ptr + length) = FREE_END;
  memset(ptr, FILL_BYTE, length);
  ptr -= sizeof(long) * 2;	/* point to malloc'd block */
  free(ptr);
}

/*
 *  Create a new duplicate copy of a string using zalloc and strcpy.
 *  This function is the same as strdup, but is included here to make
 *  use of the debugging zalloc and zfree functions. strdup itself
 *  is not part of the ANSI standard, although many compilers implement
 *  it. Here, it is implemented without relying on strdup existing.
 */
char * zstrdup(char *str)
{
  char *str2;

  str2 = zalloc(strlen(str) + 1);
  return strcpy(str2, str);
}

/*
 *  A function for checking that a string only contains whitespace
 *  and one integer, returns 0 if the string doesn't contain just
 *  an integer, or 1 if it contains optional whitespace, followed
 *  by one integer, followed by optional whitespace.
 */
int string_is_int(char *s)
{
  int state, ch;

  for (state = 1; *s != '\0'; s++) {
    ch = *s;
    if (state == 1) {
      /* expect whitespace or + or - or digit */
      if (isspace(ch))
	state = 1;
      else if (ch == '+')
	state = 2;
      else if (ch == '-')
	state = 2;
      else if (isdigit(ch))
	state = 3;
      else
	return 0;
    }
    else if (state == 2) {
      /* expect digits now */
      if (isdigit(ch))
	state = 3;
      else
	return 0;
    }
    else if (state == 3) {
      /* expect digits or whitespace or end */
      if (isdigit(ch))
	state = 3;
      else if (isspace(ch))
	state = 4;
      else
	return 0;
    }
    else if (state == 4) {
      /* expect only whitespace or end */
      if (isspace(ch))
	state = 4;
      else
	return 0;
    }
  }
  if (state >= 3)
    return 1;
  else
    return 0;
}



